@@ -5,6 +5,7 @@
 # This code is distributed under the terms and conditions
 # from the MIT License (MIT).
 #
+from collections import defaultdict
 import gzip
 import io
 import logging
@@ -12,10 +13,12 @@ import os
 import time
 import unittest
 import warnings
+from contextlib import contextmanager
+from unittest.mock import patch
 
-import boto.s3.bucket
 import boto3
 import botocore.client
+import botocore.endpoint
 import mock
 import moto
 
@@ -33,6 +36,13 @@ KEY_NAME = 'test-key'
 WRITE_KEY_NAME = 'test-write-key'
 ENABLE_MOTO_SERVER = os.environ.get("SO_ENABLE_MOTO_SERVER") == "1"
 
+#
+# This is a hack to keep moto happy
+# See https://github.com/spulec/moto/issues/1941
+#
+os.environ["AWS_ACCESS_KEY_ID"] = "test"
+os.environ["AWS_SECRET_ACCESS_KEY"] = "test"
+
 
 logger = logging.getLogger(__name__)
 
@@ -79,6 +89,26 @@ def ignore_resource_warnings():
     warnings.filterwarnings("ignore", category=ResourceWarning, message="unclosed.*<ssl.SSLSocket.*>")  # noqa
 
 
+class BaseTest(unittest.TestCase):
+    @contextmanager
+    def assertApiCalls(self, **expected_api_calls):
+        """ Track calls to S3 in self.api_calls by patching botocore.endpoint.Endpoint.make_request. """
+        _real_make_request = botocore.endpoint.Endpoint.make_request
+        api_calls = defaultdict(int)
+
+        def mock_make_request(self, operation_model, *args, **kwargs):
+            api_calls[operation_model.name] += 1
+            return _real_make_request(self, operation_model, *args, **kwargs)
+
+        patcher = patch('botocore.endpoint.Endpoint.make_request', new=mock_make_request)
+        patcher.start()
+        try:
+            yield api_calls
+            self.assertDictEqual(expected_api_calls, api_calls)
+        finally:
+            patcher.stop()
+
+
 @unittest.skipUnless(
     ENABLE_MOTO_SERVER,
     'The test case needs a Moto server running on the local 5000 port.'
@@ -97,15 +127,14 @@ class SeekableRawReaderTest(unittest.TestCase):
 
     def test_read_from_a_closed_body(self):
         obj = self._local_resource.Object(BUCKET_NAME, KEY_NAME)
-        content_length = obj.content_length
-        reader = smart_open.s3._SeekableRawReader(obj, content_length)
+        reader = smart_open.s3._SeekableRawReader(obj)
         self.assertEqual(reader.read(1), b'1')
         reader._body.close()
         self.assertEqual(reader.read(2), b'23')
 
 
 @moto.mock_s3
-class SeekableBufferedInputBaseTest(unittest.TestCase):
+class SeekableBufferedInputBaseTest(BaseTest):
     def setUp(self):
         # lower the multipart upload size, to speed up these tests
         self.old_min_part_size = smart_open.s3.DEFAULT_MIN_PART_SIZE
@@ -113,6 +142,8 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
 
         ignore_resource_warnings()
 
+        super().setUp()
+
     def tearDown(self):
         smart_open.s3.DEFAULT_MIN_PART_SIZE = self.old_min_part_size
         cleanup_bucket()
@@ -124,17 +155,19 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         put_to_bucket(contents=expected)
 
         # connect to fake s3 and read from the fake key we filled above
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        output = [line.rstrip(b'\n') for line in fin]
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
+            output = [line.rstrip(b'\n') for line in fin]
         self.assertEqual(output, expected.split(b'\n'))
 
     def test_iter_context_manager(self):
         # same thing but using a context manager
         expected = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=expected)
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            output = [line.rstrip(b'\n') for line in fin]
-            self.assertEqual(output, expected.split(b'\n'))
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
+                output = [line.rstrip(b'\n') for line in fin]
+        self.assertEqual(output, expected.split(b'\n'))
 
     def test_read(self):
         """Are S3 files read correctly?"""
@@ -142,68 +175,80 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         put_to_bucket(contents=content)
         logger.debug('content: %r len: %r', content, len(content))
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        self.assertEqual(content[:6], fin.read(6))
-        self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
-        self.assertEqual(content[14:], fin.read())  # read the rest
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
+            self.assertEqual(content[:6], fin.read(6))
+            self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
+            self.assertEqual(content[14:], fin.read())  # read the rest
 
     def test_seek_beginning(self):
         """Does seeking to the beginning of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        self.assertEqual(content[:6], fin.read(6))
-        self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
+            self.assertEqual(content[:6], fin.read(6))
+            self.assertEqual(content[6:14], fin.read(8))  # ř is 2 bytes
 
-        fin.seek(0)
-        self.assertEqual(content, fin.read())  # no size given => read whole file
+        with self.assertApiCalls(GetObject=1):
+            fin.seek(0)
+            self.assertEqual(content, fin.read())  # no size given => read whole file
 
-        fin.seek(0)
-        self.assertEqual(content, fin.read(-1))  # same thing
+        with self.assertApiCalls(GetObject=1):
+            fin.seek(0)
+            self.assertEqual(content, fin.read(-1))  # same thing
 
     def test_seek_start(self):
         """Does seeking from the start of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        seek = fin.seek(6)
-        self.assertEqual(seek, 6)
-        self.assertEqual(fin.tell(), 6)
-        self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, defer_seek=True)
+            seek = fin.seek(6)
+            self.assertEqual(seek, 6)
+            self.assertEqual(fin.tell(), 6)
+            self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
 
     def test_seek_current(self):
         """Does seeking from the middle of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        self.assertEqual(fin.read(5), b'hello')
-        seek = fin.seek(1, whence=smart_open.constants.WHENCE_CURRENT)
-        self.assertEqual(seek, 6)
-        self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
+            self.assertEqual(fin.read(5), b'hello')
+
+        with self.assertApiCalls(GetObject=1):
+            seek = fin.seek(1, whence=smart_open.constants.WHENCE_CURRENT)
+            self.assertEqual(seek, 6)
+            self.assertEqual(fin.read(6), u'wořld'.encode('utf-8'))
 
     def test_seek_end(self):
         """Does seeking from the end of S3 files work correctly?"""
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        seek = fin.seek(-4, whence=smart_open.constants.WHENCE_END)
-        self.assertEqual(seek, len(content) - 4)
-        self.assertEqual(fin.read(), b'you?')
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, defer_seek=True)
+            seek = fin.seek(-4, whence=smart_open.constants.WHENCE_END)
+            self.assertEqual(seek, len(content) - 4)
+            self.assertEqual(fin.read(), b'you?')
 
     def test_detect_eof(self):
         content = u"hello wořld\nhow are you?".encode('utf8')
         put_to_bucket(contents=content)
 
-        fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
-        fin.read()
-        eof = fin.tell()
-        self.assertEqual(eof, len(content))
-        fin.seek(0, whence=smart_open.constants.WHENCE_END)
-        self.assertEqual(eof, fin.tell())
+        with self.assertApiCalls(GetObject=1):
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME)
+            fin.read()
+            eof = fin.tell()
+            self.assertEqual(eof, len(content))
+            fin.seek(0, whence=smart_open.constants.WHENCE_END)
+            self.assertEqual(eof, fin.tell())
+            fin.seek(eof)
+            self.assertEqual(eof, fin.tell())
 
     def test_read_gzip(self):
         expected = u'раcцветали яблони и груши, поплыли туманы над рекой...'.encode('utf-8')
@@ -227,9 +272,10 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
             self.assertEqual(zipfile.read(), expected)
 
         logger.debug('starting actual test')
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            with gzip.GzipFile(fileobj=fin) as zipfile:
-                actual = zipfile.read()
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
+                with gzip.GzipFile(fileobj=fin) as zipfile:
+                    actual = zipfile.read()
 
         self.assertEqual(expected, actual)
 
@@ -237,13 +283,14 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         content = b'englishman\nin\nnew\nyork\n'
         put_to_bucket(contents=content)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            fin.readline()
-            self.assertEqual(fin.tell(), content.index(b'\n')+1)
+        with self.assertApiCalls(GetObject=2):
+            with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
+                fin.readline()
+                self.assertEqual(fin.tell(), content.index(b'\n')+1)
 
-            fin.seek(0)
-            actual = list(fin)
-            self.assertEqual(fin.tell(), len(content))
+                fin.seek(0)
+                actual = list(fin)
+                self.assertEqual(fin.tell(), len(content))
 
         expected = [b'englishman\n', b'in\n', b'new\n', b'york\n']
         self.assertEqual(expected, actual)
@@ -252,8 +299,9 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         content = b'englishman\nin\nnew\nyork\n'
         put_to_bucket(contents=content)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, buffer_size=8) as fin:
-            actual = list(fin)
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, buffer_size=8) as fin:
+                actual = list(fin)
 
         expected = [b'englishman\n', b'in\n', b'new\n', b'york\n']
         self.assertEqual(expected, actual)
@@ -262,8 +310,10 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         content = b'englishman\nin\nnew\nyork\n'
         put_to_bucket(contents=content)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            data = fin.read(0)
+        with self.assertApiCalls():
+            # set defer_seek to verify that read(0) doesn't trigger an unnecessary API call
+            with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, defer_seek=True) as fin:
+                data = fin.read(0)
 
         self.assertEqual(data, b'')
 
@@ -271,8 +321,10 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
         contents = b'the spice melange\n'
         put_to_bucket(contents=contents)
 
-        with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME) as fin:
-            returned_obj = fin.to_boto3()
+        with self.assertApiCalls():
+            # set defer_seek to verify that to_boto3() doesn't trigger an unnecessary API call
+            with smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, defer_seek=True) as fin:
+                returned_obj = fin.to_boto3()
 
         boto3_body = returned_obj.get()['Body'].read()
         self.assertEqual(contents, boto3_body)
@@ -280,10 +332,26 @@ class SeekableBufferedInputBaseTest(unittest.TestCase):
     def test_binary_iterator(self):
         expected = u"выйду ночью в поле с конём".encode('utf-8').split(b' ')
         put_to_bucket(contents=b"\n".join(expected))
-        with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'rb') as fin:
-            actual = [line.rstrip() for line in fin]
+        with self.assertApiCalls(GetObject=1):
+            with smart_open.s3.open(BUCKET_NAME, KEY_NAME, 'rb') as fin:
+                actual = [line.rstrip() for line in fin]
         self.assertEqual(expected, actual)
 
+    def test_defer_seek(self):
+        content = b'englishman\nin\nnew\nyork\n'
+        put_to_bucket(contents=content)
+
+        with self.assertApiCalls():
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, defer_seek=True)
+        with self.assertApiCalls(GetObject=1):
+            self.assertEqual(fin.read(), content)
+
+        with self.assertApiCalls():
+            fin = smart_open.s3.SeekableBufferedInputBase(BUCKET_NAME, KEY_NAME, defer_seek=True)
+        with self.assertApiCalls(GetObject=1):
+            fin.seek(10)
+            self.assertEqual(fin.read(), content[10:])
+
 
 @moto.mock_s3
 class MultipartWriterTest(unittest.TestCase):
@@ -543,18 +611,24 @@ class IterBucketTest(unittest.TestCase):
         results = list(smart_open.s3.iter_bucket(BUCKET_NAME))
         self.assertEqual(len(results), 10)
 
+    def test_deprecated_top_level_s3_iter_bucket(self):
+        populate_bucket()
+        with self.assertLogs(smart_open.logger.name, level='WARN') as cm:
+            # invoking once will generate a warning
+            smart_open.s3_iter_bucket(BUCKET_NAME)
+            # invoking again will not (to reduce spam)
+            smart_open.s3_iter_bucket(BUCKET_NAME)
+            # verify only one output
+            assert len(cm.output) == 1
+            # verify the suggested new import is in the warning
+            assert "from smart_open.s3 import iter_bucket as s3_iter_bucket" in cm.output[0]
+
     def test_accepts_boto3_bucket(self):
         populate_bucket()
         bucket = boto3.resource('s3').Bucket(BUCKET_NAME)
         results = list(smart_open.s3.iter_bucket(bucket))
         self.assertEqual(len(results), 10)
 
-    def test_accepts_boto_bucket(self):
-        populate_bucket()
-        bucket = boto.s3.bucket.Bucket(name=BUCKET_NAME)
-        results = list(smart_open.s3.iter_bucket(bucket))
-        self.assertEqual(len(results), 10)
-
     def test_list_bucket(self):
         num_keys = 10
         populate_bucket()
@@ -573,44 +647,6 @@ class IterBucketTest(unittest.TestCase):
         expected = ['key_%d' % x for x in range(num_keys)]
         self.assertEqual(sorted(keys), sorted(expected))
 
-    def test_old(self):
-        """Does s3_iter_bucket work correctly?"""
-        #
-        # Use an old-school boto Bucket class for historical reasons.
-        #
-        mybucket = boto.s3.bucket.Bucket(name=BUCKET_NAME)
-
-        # first, create some keys in the bucket
-        expected = {}
-        for key_no in range(42):
-            key_name = "mykey%s" % key_no
-            with smart_open.smart_open("s3://%s/%s" % (BUCKET_NAME, key_name), 'wb') as fout:
-                content = '\n'.join("line%i%i" % (key_no, line_no) for line_no in range(10)).encode('utf8')
-                fout.write(content)
-                expected[key_name] = content
-
-        # read all keys + their content back, in parallel, using s3_iter_bucket
-        result = {}
-        for k, c in smart_open.s3.iter_bucket(mybucket):
-            result[k] = c
-        self.assertEqual(expected, result)
-
-        # read some of the keys back, in parallel, using s3_iter_bucket
-        result = {}
-        for k, c in smart_open.s3.iter_bucket(mybucket, accept_key=lambda fname: fname.endswith('4')):
-            result[k] = c
-        self.assertEqual(result, dict((k, c) for k, c in expected.items() if k.endswith('4')))
-
-        # read some of the keys back, in parallel, using s3_iter_bucket
-        result = dict(smart_open.s3.iter_bucket(mybucket, key_limit=10))
-        self.assertEqual(len(result), min(len(expected), 10))
-
-        for workers in [1, 4, 8, 16, 64]:
-            result = {}
-            for k, c in smart_open.s3.iter_bucket(mybucket):
-                result[k] = c
-            self.assertEqual(result, expected)
-
 
 @moto.mock_s3
 @unittest.skipIf(not smart_open.concurrency._CONCURRENT_FUTURES, 'concurrent.futures unavailable')
